/*
 * HashtableTemplate.h
 *
 */

#include "LinkedListTemplate.h"
#include "HashFunctionTemplate.h"
#include <iostream>
#include <utility>


template<typename KEY, typename VALUE>
class HashTable {
public:
	HashTable(int initTableSize);
	~HashTable();

	void insert(KEY key, const VALUE &newDataItem);
	bool remove(KEY searchKey);
	bool retrieve(KEY searchKey, VALUE &dataItem);
	void clear();

	bool isEmpty() const;
	bool isFull() const;

	void printDebug() const;

private:
	int tableSize;
	List<std::pair<KEY, VALUE> > *tableData;
};

template<typename KEY, typename VALUE>
HashTable<KEY, VALUE>::HashTable(int initTableSize) :
		tableSize(initTableSize) {
	tableData = new List<std::pair<KEY, VALUE> > [tableSize];
}

template<typename KEY, typename VALUE>
HashTable<KEY, VALUE>::~HashTable() {
	delete[] tableData;
}

template<typename KEY, typename VALUE>
void HashTable<KEY, VALUE>::insert(KEY key, const VALUE &newDataItem) {
	int index = MyHash<KEY>::compute(key) % tableSize;

	if (tableData[index].isEmpty())
		tableData[index].insert(std::make_pair(key, newDataItem));
	else {
		tableData[index].gotoBeginning();
		//if there is a linked list at that index, then cycle through the linked
		//list until you reach the end of the link
		do {
			//if you find the exact same key, then replace it
			std::pair<KEY, VALUE> data;
			tableData[index].getCursor(data);
			if (data.first == key) {
				tableData[index].replace(std::make_pair(key, newDataItem));
				return;
			}
		} while (tableData[index].gotoNext());
		//then insert the data into the HashTable
		tableData[index].insert(std::make_pair(key, newDataItem));
	}
}

template<typename KEY, typename VALUE>
bool HashTable<KEY, VALUE>::remove(KEY searchKey) {
	int index = MyHash<KEY>::compute(searchKey) % tableSize;

	if (tableData[index].isEmpty())
		return false;

	tableData[index].gotoBeginning();
	do {
		if (tableData[index].getCursor().first == searchKey) {
			tableData[index].remove();
			return true;
		}
	} while (tableData[index].gotoNext());

	return false;
}

template<typename KEY, typename VALUE>
bool HashTable<KEY, VALUE>::retrieve(KEY searchKey, VALUE &dataItem) {
	int index = MyHash<KEY>::compute(searchKey) % tableSize;

	//if there is nothing at that index, then the searchkey isn't in the table
	if (tableData[index].isEmpty())
		return false;

	tableData[index].gotoBeginning();
	//cycle through the linked list comparing the names (the keys)
	do {
		if (tableData[index].getCursor().first == searchKey) {
			//a match has been found, store it in dataitem so that you
			//can compare the password information in main
			dataItem = tableData[index].getCursor();
			return true;
		}
	} while (tableData[index].gotoNext());

	return false;
}

template<typename KEY, typename VALUE>
void HashTable<KEY, VALUE>::clear() {
	for (int i = 0; i < tableSize; i++) {
		tableData[i].clear();
	}
}

template<typename KEY, typename VALUE>
bool HashTable<KEY, VALUE>::isEmpty() const {
	for (int i = 0; i < tableSize; i++) {
		if (!tableData[i].isEmpty())
			return false;
	}
	return true;
}

template<typename KEY, typename VALUE>
bool HashTable<KEY, VALUE>::isFull() const {
	for (int i = 0; i < tableSize; i++) {
		if (!tableData[i].isFull())
			return false;
	}
	return true;
}

template<typename KEY, typename VALUE>
void HashTable<KEY, VALUE>::printDebug() const {
	std::cout << "The Hash Table has the following entries" << std::endl;
	for (int i = 0; i < tableSize; i++) {
		std::cout << i << ": ";
		if (tableData[i].isEmpty())
			std::cout << "_";
		else {
			tableData[i].gotoBeginning();
			do {
				std::pair<KEY, VALUE> data;
				tableData[i].getCursor(data);
				std::cout << data.second << " ";
			} while (tableData[i].gotoNext());
		}
		std::cout << std::endl;
	}
}

